/*! \file
******************************************************************************
*----------------------------------------------------------------------------*
* Date:     28 November 2007
*
* File:     Client.cc
*
* Author:   Ed Leaver
*
* Purpose:  Listen for udp broadcast packets from Server, print received
*           messages on stdout
*           
*           
*****************************************************************************/

#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <poll.h>
#include <pwd.h>
#include <errno.h>
#include <assert.h>

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>

#include "ClientServer.hh"

using namespace std;

static string_t m_server_hostname(SERVER_HOSTNAME);
static string_t m_product_name(PRODUCT_NAME);

/// Setup the Server's listener socket to service Server ping requests
struct sigaction m_original_sigio_action_struct;
static int    m_listener_sockfd = 0;
static nfds_t m_listener_nfds   = 0;
static pollfd m_listener_fds;
static pid_t  m_pid;

bool sigio_listen_init(){
  sockaddr_in serv_addr;
  errno = 0;

  memset(&serv_addr, '\0', sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);  // i.e., server is "us"
  serv_addr.sin_port = htons(CLIENT_UDP_PORT);    // same for all us clients

  // get a socket to receive ping requests from main License Server
  if( (m_listener_sockfd=socket(AF_INET, SOCK_DGRAM, 0)) < 0){
    perror("socket");
    return false;
  }

  // tell the socket we want all us client ping-echo servers on the same machine to
  // re-use the same sock address. This way each client process will recieve a copy of
  // the ping request broadcast from the Server
  int reuse = 1;
  if(setsockopt(m_listener_sockfd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0){
    perror("setsockopt reuse:");
    return false;
  }

  // enable async IO on the socket. This means SIGIO will be sent when a new I/O event 
  // occurs. SIOCSPGRP and FIOASYNC ioctl's are both required.
  int async = 1;
  if(ioctl(m_listener_sockfd, FIOASYNC, &async) < 0){
    perror("ioctl(FIOAYNC,true):");
    return false;
  }
  // tell kernel to send the async SIGIO (same as SIGPOLL) signal to this process. 
  if(ioctl(m_listener_sockfd, SIOCSPGRP, &m_pid) < 0){
    perror("ioctl(SIOCSPGRP):");
    return false;
  }
  // bind our local server address
  if(bind(m_listener_sockfd, (sockaddr*)&serv_addr, sizeof(serv_addr)) < 0){
    perror("bind:");
    return false;
  }

  m_listener_fds.fd      = m_listener_sockfd;
  m_listener_fds.events  = POLLIN;
  m_listener_fds.revents = 0;
  m_listener_nfds = 1;
  return true;
}

/// Listen for SIGIO signals telling us m_listener_sockfd is ready to read, and
/// reply to any Server Ping requests.
void sigio_listen_handler(int sig){
  if(sig == SIGIO){
    cerr << "SIGIO received" << endl;
  }else if(sig == SIGPOLL){  // same as SIGIO
    cerr << "SIGPOLL received" << endl;
  }else if(sig == SIGUSR1){
    cerr << "SIGUSR1 received" << endl;
  }else if(sig == SIGCONT){
    cerr << "SIGCONT received" << endl;
  }else{
    cerr << "Signal " << sig << " received" << endl;
  }

  /// Note: use ppoll() in RHEL>=5
  int millisec_timeout = 1000; // since we recieved the signal, we should be good to read
  int ready = poll(&m_listener_fds, m_listener_nfds, millisec_timeout);
  if(ready < 0){
    // error, look to errno
    perror("poll");
    return;
  }else if(ready==0){
    // timed out, we're done
    cerr << "poll() timed out - shouldn't happen" << endl;
    return;
  }else{
    // only choice here is ready=1, as nfds=1 
    // look at the returned events field of our lone file descriptor
    if(m_listener_fds.revents & POLLIN){
      // data to be read, this is normal case
      /*
	}else if(m_listener_fds.revents & POLLRDHUP){
	// _GNU_SOURCE and ppoll() only, RHEL 5 and later
	#ifdef DEBUG
	cerr << "poll(): POLLRDHUP" << endl;
	#endif
	break;
      */
    }else if(m_listener_fds.revents & POLLHUP){
      cerr << "poll(): POLLHUP" << endl;
      return;
    }else if(m_listener_fds.revents & POLLERR){
      cerr << "poll(): POLLERR" << endl;
      return;
    }else if(m_listener_fds.revents & POLLNVAL){
      cerr << "poll(): POLLNVAL" << endl;
      return;
    }else{
      cerr << "poll(): revents = " << m_listener_fds.revents << endl;
      return;
    }
  }
  m_listener_fds.revents = 0; // reset for next call

  //----------------------------------------------------------------------
  // dg_echo() stuff	    (Stevens pg 282)
  const int   buflen = 1024; // 8k max for datagram, but only 1.5k over ethernet
  char rcvbuf[buflen];

  sockaddr_in cli_addr;
  socklen_t clilen = sizeof(cli_addr);

  // now read any Server ping requests and reply letting the Server know we're here

  rcvbuf[0] = '\0';

  int flags = MSG_WAITALL;
  int n = recvfrom(m_listener_sockfd, rcvbuf, buflen, flags, (sockaddr*)&cli_addr, &clilen);
  if(n<0){
    if(errno != EAGAIN){
      perror("sigio_listen_handler recvfrom:");
    }
    return;
  }else if(n==0){
    return;
  }else if(n>0){
    rcvbuf[n] = '\0';
    string_t         server_hostname = rcvbuf;
    string_t         productname     = rcvbuf+strlen(rcvbuf)+1;

    if(server_hostname == m_server_hostname && 
       productname     == m_product_name){
      // its for us, send a reply
      cout << "PID " << m_pid << ": Received ping request for "
	   <<  productname << " from server " << server_hostname << endl;
    }else{
      cout << "PID " << m_pid << ": Ping request for " << productname << " from "
	   << server_hostname << " (not for us)" << endl;
      return; // ping request not for us
    }
  } // endif (n>0)
}

/// Listen for SIGIO signals telling us m_listener_sockfd is ready to read, and
/// reply to any Client Ping requests.
void sigio_action(int sig, siginfo_t * siginfo, void * context){
  //  ucontext_t * ucontext = (ucontext_t*)context; // not used
  /// Its generally a bad idea to do much more than set a few flags in a signal handler,
  /// but we do much more anyway.
  if(sig == SIGIO){
    cerr << "SIGIO action received" << endl;
  }else if(sig == SIGPOLL){ // same as SIGIO  
    cerr << "SIGPOLL action received" << endl;
  }else if(sig == SIGUSR1){
    cerr << "SIGUSR1 action received" << endl;
  }else if(sig == SIGCONT){
    cerr << "SIGCONT action received" << endl;
  }else{
    cerr << "Signal " << sig << " action received" << endl;
  }

  {
    int serrno = siginfo->si_errno;
    int scode  = siginfo->si_code;
    /// This is just so completely, totally bogus. bits/siginfo.h gives POLL_IN = 1,
    /// but the value we recieve here is 128, way out of range of the SIG_POLL enumeration.
    /// Also, siginfo->si_fd does not equal m_listener_sockfd, which makes 
    /// sigaction() no more useful than plain old signal(). 
    /// (Well, sigaction() does provide signal masking...)
    switch(scode){
    case POLL_IN:
      cerr << "data input available" << endl;
      break;
    case POLL_OUT:
      cerr << "output buffers available" << endl;
      break;
    case POLL_MSG:
      cerr << "input message available" << endl;
      break;
    case POLL_ERR:
      cerr << "i/o error" << endl;
      break;
    case POLL_PRI:
      cerr << "high priority input available" << endl;
      break;
    case POLL_HUP:
      cerr << "service disconnected" << endl;
      break;
    default:
      cerr << "BUG: unknown siginfo code " << scode << endl;
    }
    if(1 || scode==POLL_IN || scode==POLL_MSG || scode==POLL_PRI){
      // test if(1) because scode is so meaningless. Otherwise we would just insert the
      // contents of void sigio_listen_handler(int sig) here. But we'll keep them separate
      // in case we want to set sigio_listen_handler() with signal() in main();
      sigio_listen_handler(sig);
    }
  } 

  // finish with prior action
  if(m_original_sigio_action_struct.sa_flags & SA_SIGINFO){
    cerr << " old sigaction..." << endl;
    m_original_sigio_action_struct.sa_sigaction(sig, siginfo, context);
  }else{
    if(m_original_sigio_action_struct.sa_handler == SIG_IGN){
      cerr << " old handler SIG_IGN" << endl;
    }else if(m_original_sigio_action_struct.sa_handler == SIG_DFL){
      cerr << " old handler SIG_DFL" << endl;
    }else{
      cerr << " old handler custom" << endl;
      m_original_sigio_action_struct.sa_handler(sig);
    }
  }
}



int main(int argc, char * argv[]){
  int retval = 0;

  m_pid = getpid();

  struct sigaction io_act;

  // SIGIO and SIGPOLL are the same
  io_act.sa_sigaction = sigio_action;
  io_act.sa_flags     = SA_SIGINFO;
  sigemptyset(&(io_act.sa_mask));
  sigaddset(&(io_act.sa_mask),SIGCONT);
  sigaddset(&(io_act.sa_mask),SIGPOLL);

  if(sigaction(SIGIO, &io_act, &m_original_sigio_action_struct) < 0){
    cerr << "Could not install new sigaction(SIGIO)" << endl;
    retval = 1;
    return retval;
  }
  
  if(!sigio_listen_init()){
    cerr << "Could not init listener" << endl;
    retval = 1;
    return retval;
  }

  // loop for seconds, listening for clients
  int seconds = 120; // time for server to live before exiting

  for(int i=0; i<seconds; i++){
    sleep(1);
    cout << '.' << flush;
    if((i+1)%60 == 0) cout << endl;
  }
  cout << endl;
  return retval;
}
