/*! \file
******************************************************************************
*----------------------------------------------------------------------------*
* Date:     28 November 2007
*
* File:     Server.cc
*
* Author:   Ed Leaver
*
* Purpose:  broadcast udb packet to clients print received messages (if any)
*           
*           
*****************************************************************************/

#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <poll.h>
#include <pwd.h>
#include <errno.h>
#include <assert.h>

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>

#include "ClientServer.hh"

using namespace std;

/// Server Ping-Echo Client (SPE Clnt):
/// Broadcast "Ping" Request to all running clients asking they identify themselves:
/// hostname, username, and product id with version number and client starttime.
///    - get unique udp port from system
///    - create socket and bind to this udp port
///    - broadcast (or multicast when allowed config_file) ping request on socket
///      ping request should have TTL (datagram time-to-live) set to one or two seconds,
///
/// Reference: Stevens pg 288
bool PingAllClients(const string_t & ProductName,
		    const string_t & RequestingHostname){
  // init return values
  bool retval = false;

  char hostname[128];
  gethostname(hostname,128);

  const int buflen = 1024; // 8k max for datagram, but only 1.5 k over ethernet
  char msgbuf[buflen];
  memset(msgbuf,'\0',buflen);
  // put something useful in buffer
  strcpy(msgbuf,hostname);
  strcpy(msgbuf+strlen(hostname)+1,ProductName.c_str());
  size_t msglen = strlen(hostname) + ProductName.length() + 2;

  int         sockfd;
  sockaddr_in serv_addr;

  memset(&serv_addr, '\0', sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;

  // translate a "dotted-decimal" netmask 
  if(inet_aton(NETMASK, &serv_addr.sin_addr) == 0){
      perror("inet_aton");
#ifdef DEBUG
      cerr << "Error: could not set netmask, using broadcast" << endl;
#endif
#ifdef FEDORA8
      serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);  // Works on F8 and CentOS-5, but not on CentOS-4
#else
      serv_addr.sin_addr.s_addr = htonl(INADDR_BROADCAST); // works on CentOS-4 and 5, but not Fedora 8
#endif
  }
  serv_addr.sin_port = htons(CLIENT_UDP_PORT); // Common port number for all clients

  // get a socket to send ping request to remote ping servers
  if( (sockfd=socket(PF_INET, SOCK_DGRAM, 0)) < 0){
    perror("socket");
    exit(1);
  }
  // tell the socket we want to broadcast on it
  int bcast = 1;
  if(setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, &bcast, sizeof(bcast)) < 0){
    perror("setsockopt");
  }
  // tell the socket not to let the datagram live too long, since our clients'
  // ping-echo servers might only peek at the datagram, not actually receive it
  // (man 7 ip)
  int TimeToLive = 1; // set larger only if you want packet to leave local network
  if(setsockopt(sockfd, IPPROTO_IP, IP_TTL, &TimeToLive, sizeof(TimeToLive)) < 0){
    perror("setsockopt IP_TTL");
  }
  TimeToLive = 1; // set larger only if you want packet to leave local network
  if(setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL, &TimeToLive, sizeof(TimeToLive)) < 0){
    perror("setsockopt IP_MULTICAST_TTL");
  }
  
  // broadcast the message
  if(sendto(sockfd, msgbuf, msglen, 0, (sockaddr*)&serv_addr, sizeof(serv_addr)) != msglen){
    perror("sendto");
    return (retval=false);
  }
  cout << "Broadcast Message: " << msgbuf << ' ' << (msgbuf+strlen(hostname)+1) << endl;
  return (retval=true);
}

int main(int argc, char * argv[0]){
  string_t server_hostname(SERVER_HOSTNAME);
  string_t product_name(PRODUCT_NAME);
  if(PingAllClients(product_name,server_hostname)){
    return 0;
  }else{
    return 1;
  }
}
